function [X, info] = globalnewtonlike(f, gradf, Mkf, x0, tol, maxit, beta, gamma, a1, a2, p)
% X = globalnewtonlike(f, gradf, Mkf, x0, tol, maxit, beta, gamma, a1, a2, p)
%
% Locates a stationary point of function f using a globalized Newton-like method with
% Armijo backtracking step size strategy and steepest descent fallback.
%
% INPUT:        f --> function handle to objective             (accepting one input variable)
%           gradf --> function handle to objective's gradient  (accepting one input variable)
%             Mkf --> function handle to hessian approximation (accepting one input variable)
%              x0 --> initial guess (starting point)
%             tol --> stopping condition tolerance:   ||gradf|| <= tol
%           maxit --> maximum number of iterations
%            beta --> Armijo parameter (step size shrinking)
%           gamma --> Armijo parameter (sufficient descent)
%       a1, a2, p --> parameters for direction selection
%
% OUTPUT: 
%          X --> matrix with all iterates; X(:,k) containing the k-th iterate
%       info --> -1, if maximum number of iterations is reached
%                 0, if a stationary point (up to tol) is found
%


% ensure initial guess x0 is a single column
x0 = reshape(x0, [], 1);

% setup and defaults
info     = -1;        % error status value: maximum number of iterations reached
xk       = x0;        % initial guess is zero-th iterate 

% prepare output (NOTE: Matlab uses column-major order!)
historyChunk = 1000;
dimx = length(x0);
X = zeros(dimx, historyChunk);

% store initial guess
X(:,1) = x0;
historyIdx  = 1;  % fill counter
historySize = historyChunk;

% compute first gradient at xk
gradf_xk = gradf(xk);

% initial output
fprintf('..');  % for formatted output
outputIteration(0, NaN, f, X, historyIdx, NaN, gradf_xk)

% iterate
for k = 1:maxit

   % check the stopping criterion 
   if norm(gradf_xk,2) <= tol
      info = 0; % set output flag
      break;
   end
   
   % evaluate hessian approximation at xk
   Mk = Mkf(xk);
   
   % try solving Newton system
   [dkN, R] = linsolve(Mk, -gradf_xk);
   
   % check condition of hessian; if too bad, do not use the Newton direction
   if (R < 1.0d-10)
      dkN = 0;
   end
 
   % select search direction according to direction condition (7.1)
   if (norm(dkN,2) >= 1.0d-15) && ( -gradf_xk.' * dkN  >=  min(a1, a2*norm(dkN,2)^p) * norm(dkN,2)^2 )
      dk = dkN;
      fprintf('N'); % display N for newton direction
   else
      dk = - gradf_xk;
      fprintf('G'); % display G for gradient direction
   end
   
   % determine Armijo step size in direction dk
   sigmak = armijo(f, gradf_xk, xk, dk, gamma, beta, 0);   
   
   % perform the step
   xk = xk + sigmak * dk;
   
   % store it in the result matrix; grow-and-copy only if necessary
   historyIdx = historyIdx + 1;
   if historyIdx > historySize
      X(:,end+1:end+historyChunk) = 0;          % increase history size
      historySize = historySize + historyChunk; % new size of history
   end
   X(:,historyIdx) = xk;
   
   % compute gradient at xk for next iteration
   gradf_xk = gradf(xk);
   
   % output
   outputIteration(k, R, f, X, historyIdx, sigmak, gradf_xk)
     
end

% shrink result array to appropriate size
X = X(:, 1:historyIdx);

% finito of function
return

end % end of function



% HELPER FOR OUTPUT
function outputIteration(k, R, f, X, historyIdx, sigmak, gradfxk)
   % INPUT:  k --> iteration number
   %         R --> reciprocal condition of matrix Mk
   %         f --> function handle to objective f
   %         X --> history storage
   %      xkm1 --> previous iterate
   %      xkm2 --> penultimate iterate
   %    sigmak --> step size
   %   gradfxk --> gradient at xk
   
   % only output for 2d
   if length(X(:,1)) ~= 2, return; end
   
   % compute/set some intermediate
   xk = X(:, historyIdx);                                     % xk   = current iterate
   if k>1, xkm1 = X(:, historyIdx-1); else, xkm1 = NaN; end   % xkm1 = previous iterate
   if k>2, xkm2 = X(:, historyIdx-2); else, xkm2 = NaN; end   % xkm2 = penultimate iterate
   progress = norm(xk-xkm1,2) / norm(xkm1-xkm2,2);
   
   % output
   fprintf(' #%3d: ', k);  % iteration
   fprintf('R=%11.5g', R); % reciprocal hessian condition
   fprintf('| f(x)=%11.5g ', f(xk));  % function value
   fprintf('| x(1)=%11.5g | x(2)=%11.5g ', xk(1), xk(2));   % current iterate
   %fprintf('| d(1)=%11.5g | d(2)=%11.5g ', dk(1), dk(2));  % search direction
   fprintf('| step=%11.5g   ', sigmak);   % step size
   fprintf('| crate=%11.5g ', progress);           % progress measure: contraction rate
   fprintf('| ||gradf||=%11.5g ', norm(gradfxk));  % norm of gradient
   fprintf('\n');  % new line
end
