function [fv, f, gradf, hessf, x0, xout] = testFun()

% Rosenbrock: objective and its gradient
% fv = @(x1,x2) 100 * ( x2  -  x1.^2).^2 + (1 -  x1 ).^2;                               % vectorized for fast visualization
% f  = @(x)     100 * (x(2) - x(1)^2)^2  + (1 - x(1))^2;                                % non-vectorized with vector input
% gradf = @(x)  [ -400*x(1)*x(2) + 400*x(1)^3 + 2*x(1) - 2   ;   200*(x(2)-x(1)^2) ];   % non-vectorized with vector input
% hessf = @(x)  [ -400*x(2)+1200*x(1)^2 + 2   ,   -400*x(1)  ; ... 
%                 -400*x(1)                   ,   200       ];

% Test function
% fv    = @(x1,x2) (0.5*x1-0.2).^4 + (x2-0.6).^2        ;
% f     = @(x) (0.5*x(1)-0.2)^4 + (x(2)-0.6)^2          ;  % minimum at x = [0.4 ; 0.6]
% gradf = @(x) [4*0.5*(0.5*x(1)-0.2)^3  ; 2*(x(2)-0.6)] ;
% hessf = @(x) [    3*(0.5*x(1)-0.2)^2  ,  0            ; ...
%                                    0  ,  2           ];
                                
% Test function set
% fv    = @(x1,x2) (x1.*x2.^2 - x1 + 9/4).^2 + (x1.*x2.^3 - x1 + 21/8).^2 + (x1.*x2 - x1 + 3/2).^2;

% Test function: Powell's function
% fv    = @(x1,x2,x3,x4) (x1+10*x2).^2 + 5*(x3-x4).^2 + (x2-2*x3).^4 + 10*(x1-x4).^4;

% Test function
% fv = @(x1, x2) (x1+x2).^2 .* (4*x1.^2 + 2*x2.^2 + 4*x1.*x2 + 2*x1 + 1);

% Test function:  Exponential
fv = @(x1,x2) log(exp(x1) + exp(-x1)) + x2.^2;

% Test function: Himmelblau's function
% fv = @(x1, x2) (x1.^2 + x2 - 11).^2 + (x1 + x2.^2 - 7).^2;



% auto-generate f, gradf, hessf using symbolic notations
xsymb = num2cell(sym('x',[1 nargin(fv)]));
fvsym = fv(xsymb{:});
f     = matlabFunction(         fvsym , 'Vars', {[xsymb{:}].'});
gradf = matlabFunction(gradient(fvsym), 'Vars', {[xsymb{:}].'});
hessf = matlabFunction( hessian(fvsym), 'Vars', {[xsymb{:}].'});

                                
% initial points
x0_a = [  1.0 ;  -0.5]; 
x0_b = [ -1.2 ;   1.0]; 
x0_c = [  5.0 ;   5.0];
%x0_d = [ 33.0 ; -62.0];
%x0_powell = [ 3 ; -1 ; 0 ; 1 ]; % for Powell's function


% optimization settings
tol   = 1.0d-6;   % stopping criterion || gradf || <= tol
maxit = 100;      % maximum number of iterations 
beta  = 0.5;      % Armijo step size shrink factor
gamma = 1.0d-4;   % Armijo sufficient descent parameter
a1    = 1d-6;     %
a2    = 1d-6;     % Newton step acceptance parameters
p     = 0.1;      %


% Select starting point
x0 = x0_a;
%x0 = x0_b;
%x0 = x0_c;

% Set hessian approximations
hessfXk     = @(x) hessf(x);              % exact hessian
hessfXkDiag = @(x) diag(diag(hessf(x)));  % exact hessian diagonal: INEFFICIENT EVALUATION! (but convenient)
hessfX0     = @(x) hessf(x0);             % always use initial hessian

% Optimize and store in output variable
xout.hessfXk     = optimizeWith(x0, hessfXk);      fprintf('\n\n\n%s\n\n\n', repmat('=', 1, 100));
xout.hessfXkDiag = optimizeWith(x0, hessfXkDiag);  fprintf('\n\n\n%s\n\n\n', repmat('=', 1, 100));
xout.hessfX0     = optimizeWith(x0, hessfX0);      fprintf('\n\n\n%s\n\n\n', repmat('=', 1, 100));

% finito
return





% helper function
function xend = optimizeWith(x0, Mk)

   % display info
   fprintf('\nMk function: %s', func2str(Mk));
   fprintf('\nx0 = [%.2f ; %.2f]\n', x0(1), x0(2));
   
   % augment function with an evaluation counter (slow, but just for the statistics)
   af     = augmentWithEvalCounter(f);
   agradf = augmentWithEvalCounter(gradf);
   ahessf = augmentWithEvalCounter(Mk);

   % reset evaluation counters
   af('reset'); agradf('reset'); ahessf('reset');
   
   % optimize
   [X, info] = globalnewtonlike(af, agradf, ahessf, x0, tol, maxit, beta, gamma, a1, a2, p);
   %[X, info] = localnewton(af, agradf, ahessf, x0, tol, maxit);
   
   % retrieve count of function evaluations
   fn = af('count'); gradfn = agradf('count'); hessfn = ahessf('count');

   % display info
   Mkxk    = Mk(X(:,end));    % hessian approximation at solution point / last iterate
   Hessfxk = hessf(X(:,end)); % exact hessian at solution point / last iterate
   eigMkxk    = eig(Mkxk);    
   eigHessfxk = eig(Hessfxk); % eigenvalues of hessian
   fprintf('\nMk function: %s\n', func2str(Mk));
   fprintf('Hessian approximation at last iterate (rank = %d, condition = %d, eigenvalues = (%g, %g)):\n', rank(Mkxk), cond(Mkxk), eigMkxk(1), eigMkxk(2));
   disp(Mkxk);
   fprintf('Hessian (exact) at last iterate (rank = %d, condition = %d, eigenvalues = (%g, %g)):\n', rank(Hessfxk), cond(Hessfxk), eigHessfxk(1), eigHessfxk(2));
   disp(Hessfxk);
   fprintf('  x0 = [%.2f ; %.2f] \n', x0(1), x0(2));
   fprintf('   x = [%.8f ; %.8f] \n', X(1,end), X(2,end));
   fprintf('f(x) = %.8g \n', f(X(:,end)));
   fprintf(' #it = %d  #f = %d  #gradf = %d  #hessf = %d \n', size(X,2)-1, fn, gradfn, hessfn);
   fprintf('');
   
   % return value
   xend = X(:,end);
end

end
